@@ -60,7 +60,7 @@ sub _connect {
     $options{LocalAddr} = $args->{local_address} if $args->{local_address};
     $options{PeerAddr} =~ s/[\[\]]//g if $options{PeerAddr};
     my $class = IPV6 ? 'IO::Socket::IP' : 'IO::Socket::INET';
-    return $self->emit(error => "Couldn't connect: $@")
+    return $self->emit(error => "Can't connect: $@")
       unless $self->{handle} = $handle = $class->new(%options);
   }
   $handle->blocking(0);
